//===- JITScopDetection.h - Process ProfileInfo-------------------*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// Use profiling information to calculate interesting numbers for SCoPs.
// Possible numbers are for example:
//   SCoP iterations
//   SCoP coverage (BasicBlocks)
//   SCoP inner loop iterations
// This values become more useful when used in a JIT context, applied on
// rejected SCoP candidates.
//
//===----------------------------------------------------------------------===//
#ifndef POLLY_JIT_SCOP_DETECTION_H
#define POLLY_JIT_SCOP_DETECTION_H

#include "polly/LinkAllPasses.h"
#include "polly/ScopDetection.h"
#include "polly/ScopInfo.h"
#include "polly/Support/ScopHelper.h"

#include "llvm/Analysis/AliasAnalysis.h"
#include "llvm/Analysis/AliasSetTracker.h"
#include "llvm/Analysis/LoopInfo.h"
#include "llvm/Analysis/ScalarEvolution.h"
#include "llvm/Analysis/ScalarEvolutionExpressions.h"

#include "llvm/Pass.h"
#include "llvm/ADT/DenseMap.h"
#include "llvm/Support/Debug.h"

#include <vector>

using namespace llvm;
using namespace polly;

namespace llvm {
  class AliasAnalysis;
  class Region;
  class RegionInfo;
}

namespace polly {

/// Check if a given SCEV becomes affine if parameters
/// get substituted at run time.
//@{
class JITableSCEVChecker : SCEVVisitor<JITableSCEVChecker, bool> {
  friend struct SCEVVisitor<JITableSCEVChecker, bool>;

  const Region *R;
  ScalarEvolution *SE;
private:
  typedef SCEVNAryExpr::op_iterator scev_op_it;

  // DO NOT IMPLEMENT
  JITableSCEVChecker(const JITableSCEVChecker &);
  // DO NOT IMPLEMENT
  const JITableSCEVChecker &operator=(const JITableSCEVChecker &);

  bool visitConstant(const SCEVConstant *S)  {
    // We're always fine with constant expressions.
    return true;
  }

  bool visitUnknown(const SCEVUnknown* S)  {
    Value *V = S->getValue();

    Type *AllocTy;
    Constant *FieldNo;
    // We treat these as constant.
    if (S->isSizeOf  (AllocTy) ||
        S->isAlignOf (AllocTy) ||
        S->isOffsetOf(AllocTy, FieldNo))
      return true;

    // Invariant only if not contained inside the region.
    if (Instruction *I = dyn_cast<Instruction>(V)) {
      if (!R->contains(I)) {
        return true;
      } else {
        DEBUG(dbgs() << "\n Region-variant: " << *I);
      }
    }

    return false;
  }

  bool visitNAryExpr(const SCEVNAryExpr* S)  {
    for (scev_op_it I = S->op_begin(), E = S->op_end(); I != E; ++I)
      if (!visit(*I))
        return false;

    return true;
  }

  bool visitMulExpr(const SCEVMulExpr* S) {
    return visitNAryExpr(S);
  }

  bool visitCastExpr(const SCEVCastExpr *S)  {
    return visit(S->getOperand());
  }

  bool visitTruncateExpr(const SCEVTruncateExpr *S)  {
    return visit(S->getOperand());
  }

  bool visitZeroExtendExpr(const SCEVZeroExtendExpr *S)  {
    return visit(S->getOperand());
  }

  bool visitSignExtendExpr(const SCEVSignExtendExpr *S)  {
    return visit(S->getOperand());
  }

  bool visitAddExpr(const SCEVAddExpr *S)  {
    return visitNAryExpr(S);
  }

  bool visitAddRecExpr(const SCEVAddRecExpr *S)  {
    bool ret;

    // {a, +, b, +, c} Is always bad.
    // {a,+,{b ,+,c}} Is always bad.
    // {{a,+,b} ,+,c} Is always bad.
    if (isa<SCEVAddRecExpr>(S->getStepRecurrence(*SE)))
      return false;

    // Check for invariance.
    ret = visitNAryExpr(S);

    return ret;
  }

  bool visitUDivExpr(const SCEVUDivExpr *S)  {
    return visit(S->getLHS()) && visit(S->getRHS());
  }

  bool visitSMaxExpr(const SCEVSMaxExpr *S)  {
    return visitNAryExpr(S);
  }

  bool visitUMaxExpr(const SCEVUMaxExpr *S)  {
    return visitNAryExpr(S);
  }

  bool visitCouldNotCompute(const SCEVCouldNotCompute *S) {
    return false;
  }

public:
  explicit JITableSCEVChecker(const Region *r, ScalarEvolution *se) :
    R(r), SE(se) {}

  static bool isJITable(const SCEV *s, const Region *r, ScalarEvolution *se) {
    bool ret = false;
    DEBUG(dbgs() << *s);
    JITableSCEVChecker check(r,se);
    ret = check.visit(s);
    DEBUG(dbgs()  << " [ " << ((ret) ? "TRUE" : "FALSE") << " ]\n");
    return ret;
  }
};
//@}

/// Detailled information about rejected regions.
//@{
struct RejectedInfo {
  uint64_t NonAffineLoopBound;
  uint64_t NonAffineAccess;
  uint64_t NonAffineCondition;
  uint64_t Unknown;
  uint64_t TotalRejects;

  RejectedInfo() :
    NonAffineLoopBound(0),
    NonAffineAccess(0),
    NonAffineCondition(0),
    Unknown(0),
    TotalRejects(0) {}
};
//@}

/// Detailled statistics for jitable SCoPs.
//@{
struct JitableInfo {
  uint64_t TotalJitableRegions;

  JitableInfo() :
    TotalJitableRegions(0) {}
};
//@}

class JITScopDetection : public FunctionPass {
public:
  static char ID;

  explicit JITScopDetection() : FunctionPass(ID) {}

  virtual void getAnalysisUsage(AnalysisUsage &AU) const {
    AU.addRequired<RegionInfo>();
    AU.addRequired<AliasAnalysis>();
    AU.addRequired<ScalarEvolution>();
    AU.addRequired<ScopDetection>();
    AU.setPreservesAll();
  }

  virtual bool runOnFunction(Function &F);

  typedef ScopDetection::iterator iterator;

  ScopDetection::rejected_iterator rej_begin() const { return SD->rej_begin(); }
  ScopDetection::rejected_iterator rej_end() const { return SD->rej_end(); }

  iterator begin() const { return ValidRegions.begin(); }
  iterator end() const { return ValidRegions.end(); }
  iterator find(const Region *R) const {
    return ValidRegions.find(R);
  }

  RegionInfo *getRegionInfo() { return RI; }
  RejectedInfo getRejects() const { return Rejects; }
  JitableInfo getJitables() const { return Jitables; }
private:
  // Remember the valid regions
  typedef std::set<const Region*> RegionSet;
  typedef std::set<const Region*> JitableRegions;
  typedef std::set<const Region*> DepthElem;
  typedef std::vector<DepthElem> DepthList;

  JitableRegions Regions;
  ScopDetection *SD;
  RegionInfo *RI;
  ScalarEvolution *SE;
  AliasAnalysis *AA;

  void findJitableRegions(Function &F);

  DepthList initJitableRegionQueue();
  int getSubRegionsFor(Region *R);
  int getSubSCoPsFor(Region *R);
  JITScopDetection::DepthElem getNewJitableElem(
  JITScopDetection::DepthElem &elem, DepthList &levels, int level);
  void initNextLevel(DepthElem *fix, DepthElem *level, DepthList &levels);

  bool isJitableRegion(const Region *R, const RejectInfo *RI);
  void collectCoverage(Function &F);
  void collectRejectedInfo(Function &F);

  RegionSet ValidRegions;
  JitableInfo Jitables;
  RejectedInfo Rejects;
};
}

namespace llvm {
  class PassRegistry;
  void initializeJITScopDetectionPass(llvm::PassRegistry&);
}

#endif //POLLY_SCOP_PROFILE_INFO_H
