//===- ScopProfileInfo.h - Process ProfileInfo-------------------*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// Use profiling information to calculate interesting numbers for SCoPs.
// Possible numbers are for example:
//   SCoP iterations
//   SCoP coverage (BasicBlocks)
//   SCoP inner loop iterations
// This values become more useful when used in a JIT context, applied on
// rejected SCoP candidates.
//
//===----------------------------------------------------------------------===//
#ifndef POLLY_SCOP_PROFILE_INFO_H
#define POLLY_SCOP_PROFILE_INFO_H

#include "polly/LinkAllPasses.h"
#include "polly/JITScopDetection.h"
#include "polly/ScopInfo.h"
#include "polly/Support/ScopHelper.h"

#include "llvm/Analysis/AliasAnalysis.h"
#include "llvm/Analysis/AliasSetTracker.h"
#include "llvm/Analysis/ProfileInfo.h"
#include "llvm/Analysis/LoopInfo.h"
#include "llvm/Analysis/ScalarEvolution.h"
#include "llvm/Analysis/ScalarEvolutionExpressions.h"

#include "llvm/Pass.h"
#include "llvm/ADT/DenseMap.h"

#include <vector>

using namespace llvm;
using namespace polly;

namespace llvm {
  class AliasAnalysis;
  class ProfileInfoLoader;
  class Region;
  class RegionInfo;
}

namespace polly {

class ScopProfileInfo : public FunctionPass {
public:
  static char ID;

  explicit ScopProfileInfo()
    : FunctionPass(ID) {
    TotalDynExecutions = 0;
    TotalBasicBlocks = 0;
    ExecBasicBlocks = 0;
    ExecInstructions = 0;
  }

  virtual void getAnalysisUsage(AnalysisUsage &AU) const {
    AU.setPreservesAll();
    AU.addRequired<ProfileInfo>();
    AU.addRequired<RegionInfo>();
    AU.addRequired<AliasAnalysis>();
    AU.addRequired<ScalarEvolution>();
    AU.addRequired<JITScopDetection>();
  }

  virtual bool runOnFunction(Function &F);
  virtual void print(raw_ostream &OS, const Module *)  const ;

  typedef ScopDetection::iterator iterator;
  JITScopDetection::iterator begin() const { return SD->begin(); }
  JITScopDetection::iterator end() const { return SD->end(); }

  typedef std::vector<const Region *>::iterator region_iterator;
  region_iterator skipped_begin() { return SkippedScops.begin(); }
  region_iterator skipped_end() { return SkippedScops.end(); }

  RejectedInfo getRejects() const { return Rejects; }
  JitableInfo getJitables() const { return Jitables; }

  double getExecInstructions() const { return ExecInstructions; }

  /// Get the number of SCoPs that has not been
  /// executed during runtime.
  double getSkippedSCoPs() const { return SkippedScops.size(); }

  /// Get the number of instruction executions.
  double getTotalExecutions() const { return TotalDynExecutions; }

  /// Get the number of instruction executions for a given region.
  /// @param Region
  double getDynExecutions(const Region *R) const {
    ScopValueMap::const_iterator i = DynExecCounts.find(R->getNameStr());
    ScopValueMap::const_iterator e = DynExecCounts.end();
    if (i != e)
      return (*i).second;
    else
      return 0;
  }

  /// Get the number of BasicBlocks contained within the region.
  /// @param Region
  double getBasicBlockCount(const Region *R) { return BBCounts[R->getNameStr()]; }

  /// Get the number of BasicBlocks contained within the region,
  /// if the region has been executed.
  /// @param Region
  double getBasicBlockCountIfExecuted(const Region *R) {
      return (PI->getExecutionCount(R->getEntry()) > 0) ?
                  getBasicBlockCount(R) : 0;
  }

  /// Get the number of BasicBlocks that have been executed.
  double getExecutedBasicBlockCount() const { return ExecBasicBlocks; }

  /// Get the total number of BasicBlocks visited.
  double getTotalBasicBlockCount() const { return TotalBasicBlocks; }

  /// Get the number of instructions inside a SCoP.
  double getInstsInSCoP(const Region *R) const;

  /// Get the number of instructions inside an executed SCoP.
  double getInstsInExecSCoP(const Region *R) const {
        return (PI->getExecutionCount(R->getEntry()) > 0)    ?
                    getInstsInSCoP(R) : 0;
  }

  long long getRegionRuntime(const Function *F, const Region *R) const;

private:
  typedef std::map<std::string, double> ScopValueMap;
  typedef std::set<const Region*> JitableRegions;
  typedef std::set<const Region *> DepthElem;
  typedef std::vector<DepthElem> DepthList;

  JitableRegions Regions;

  ProfileInfo *PI;
  JITScopDetection *SD;
  RegionInfo *RI;
  ScalarEvolution *SE;
  AliasAnalysis *AA;

  // Remember the valid regions
  typedef std::set<const Region*> RegionSet;

  void accumulateDynamicFunctionInfo(Function &F);
  void accumulateDynamicRegionInfo(const Region *R);

  void printJitable_Regions(ScopProfileInfo::DepthElem &elem, int level);
  void printRejectedInfo(llvm::raw_ostream &OS) const;

  void collectCoverage(Function &F);
  void collectRejectedInfo(Function &F);
  void printRegionInfo(raw_ostream &OS, const Region *R,
                       const RejectInfo *RI) const;

  double TotalDynExecutions;
  double TotalBasicBlocks;
  double TotalInstructions;
  double ExecBasicBlocks;
  double ExecInstructions;
  ScopValueMap ExecCounts;
  ScopValueMap DynExecCounts;
  ScopValueMap BBCounts;
  std::vector<const Region*> SkippedScops;
  JitableInfo Jitables;
  RejectedInfo Rejects;
};
}

namespace llvm {
  class PassRegistry;
  void initializeScopProfileInfoPass(llvm::PassRegistry&);
}

#endif //POLLY_SCOP_PROFILE_INFO_H
