//===- JITScopDetection.cpp -                    ----------------*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
//
//===----------------------------------------------------------------------===//
#include "polly/JITScopDetection.h"

#include "llvm/ADT/Statistic.h"
#include "llvm/Analysis/AliasAnalysis.h"
#include "llvm/Analysis/RegionPass.h"
#include "llvm/Analysis/RegionInfo.h"
#include "llvm/Analysis/RegionIterator.h"

#include "llvm/Support/CommandLine.h"
#include "llvm/Support/raw_ostream.h"

#include "llvm/Support/Debug.h"
#include "llvm/Module.h"

#include <numeric>
#include <string>

using namespace llvm;
using namespace polly;

static cl::opt<bool>
EnableJitable("jitable", cl::desc("Add jitable SCoP coverage."),
           cl::value_desc("Add jitable SCoP coverage."),
           cl::init(false));

static cl::opt<bool>
EnableExtended("extended", cl::desc("Overapproximate jitable regions."),
           cl::value_desc("Overapproximate jitable regions."),
           cl::init(false));

STATISTIC(JITScops, "Number of jitable scops");
STATISTIC(SubScops, "Number of function calls");

bool JITScopDetection::isJitableRegion(const Region *R, const RejectInfo *RI) {
  if (!R) return false;

  bool isJitable = false;

  if ((RI->Reason == NonAffineLoopBound) ||
      (RI->Reason == NonAffineAccess) ||
      (RI->Reason == NonAffineCondition)) {

    // Query the JitableSCEV visitor to check if we can fix the
    // non-linearity later at run time.
    if (RI->Failed_LHS)
      isJitable = JITableSCEVChecker::isJITable(RI->Failed_LHS, R,
                                      polly::JITScopDetection::SE);

    if (isJitable && RI->Failed_RHS)
      isJitable = JITableSCEVChecker::isJITable(RI->Failed_RHS, R,
                                      polly::JITScopDetection::SE);

    // Assume we can handle every expression.
    if (EnableExtended)
      isJitable = true;
  }

  // Aliasing can be resolved at run time.
  if (RI->Reason == Alias)
    isJitable = true;

  // Scalar dependencies can be modeled using an array.
  if (RI->Reason == Scalar)
    isJitable = true;

  /* Check if we could have a function call inside the rest of this region.
   * Exclude this scenario for now, if we cannot prove that it is side-effect
   * free.*/
  if (isJitable) {
    for (Region::const_block_iterator i = R->block_begin(), e = R->block_end();
         i != e; ++i)
    {
      for (BasicBlock::const_iterator ii = (**i).begin(), ie = (**i).end();
           ii != ie; ++ii) {

        if (const CallInst *call = dyn_cast<const CallInst>(&*ii))
          ++SubScops;
        //  isJitable = isJitable && call->doesNotAccessMemory();
        if (const InvokeInst *invoke = dyn_cast<const InvokeInst>(&*ii))
          ++SubScops;
        //  isJitable = isJitable && invoke->doesNotAccessMemory();
      }
    }
  }

  return isJitable;
}

void JITScopDetection::initNextLevel(DepthElem *fix, DepthElem *level, DepthList &levels) {
  ScopDetection::RejectedRegionMap map = SD->getRejects();
  DepthElem next;
  DepthElem rej;

  levels.push_back(*fix);
  std::for_each(level->begin(), level->end(), [&] (const Region *r)
  {
    std::for_each(r->begin(), r->end(), [&, this] (const Region *i)
    {
      if (map.count(i)) {
        RejectInfo ri = map[i];

        if (isJitableRegion(i, &ri))
          rej.insert(i);
      }
      next.insert(i);
    });
  });

  if (!next.empty())
    initNextLevel(&rej, &next, levels);
}

JITScopDetection::DepthList JITScopDetection::initJitableRegionQueue() {
  const Region *TopRegion = RI->getTopLevelRegion();
  DepthList depth; // Contains the jitable regions sorted by the tree level.
  DepthElem rej;
  DepthElem next;

  next.insert(TopRegion);
  initNextLevel(&rej, &next, depth);
  return depth;
}

int JITScopDetection::getSubRegionsFor(Region *R) {
  int childrenCount = 0;
  // Count all subregions this parent has.
  std::for_each(R->element_begin(), R->element_end(), [&, this]
                (Region::element_iterator::pointer node)
  {
    if (R->isSubRegion())
      childrenCount++;
  });

  return childrenCount;
}

/*
 * Get the number of Subregions which are valid SCoPs.
 */
int JITScopDetection::getSubSCoPsFor(Region *R) {
  const ScopDetection::RejectedRegionMap map = SD->getRejects();
  Region *Sub;
  int scopCount = 0;

  std::for_each(R->element_begin(), R->element_end(), [&, this]
                (Region::element_iterator::pointer node)
  {
    if (node->isSubRegion()) {
      Sub = node->getNodeAs<Region>();

      if (!map.count(Sub)) {
        ScopDetection::DetectionContext Context(*Sub, *AA, false);
        if (SD->isValidRegion(Context))
          scopCount++;
      }
    }
  });
  return scopCount;
}

JITScopDetection::DepthElem JITScopDetection::getNewJitableElem(
  JITScopDetection::DepthElem &elem, DepthList &levels, int level) {

  DepthElem newElem;
  Region *Parent;
  int childrenCount, childrenFound;
  const RejectInfo *RInfo;

  std::for_each(elem.begin(), elem.end(), [&, this] (const Region * R)
  {
    Parent = R->getParent();

    childrenFound = 0;
    if (Parent && (level > 0) && !levels[level].count(Parent)) {
      // Get childrenCount of this parent.
      childrenCount  = getSubRegionsFor(Parent);
      childrenFound  = getSubSCoPsFor(Parent);

      // Count all jitable regions with the same parent.
      std::for_each(elem.begin(), elem.end(), [&, this] (const Region *R)
      {
        if (Parent == R->getParent())
          childrenFound++;
      });

      if (childrenCount > childrenFound) {
        newElem.insert(R);
      } else if (childrenCount == childrenFound) {
        // All children jitable, check if parent is jitable and insert him.
        ScopDetection::DetectionContext Context(*Parent, *AA, false);
        if (SD->isValidRegion(Context))
          llvm_unreachable("Detecting a SCoP among the rejects!");

        RInfo = &(Context.RI);
        if (isJitableRegion(Parent, RInfo) && (level > 0))
          levels[level-1].insert(Parent);
        else
          newElem.insert(R);
      }
    }
  });

  return newElem;
}

void JITScopDetection::collectRejectedInfo(Function &F) {
  // Accumulate data about rejected regions during SCoP detection.
  RejectInfo *RI;
  for (ScopDetection::rejected_iterator SI = SD->rej_begin(),
       SE = SD->rej_end(); SI != SE; ++SI) {
    RI = &((*SI).second);

    if (RI->Reason == NonAffineLoopBound)
      ++(Rejects.NonAffineLoopBound);

    if (RI->Reason == NonAffineAccess)
      ++(Rejects.NonAffineAccess);

    if (RI->Reason == NonAffineCondition)
      ++(Rejects.NonAffineCondition);

    if (RI->Reason == Unknown)
      ++(Rejects.Unknown);

    ++(Rejects.TotalRejects);
  }
}

void JITScopDetection::collectCoverage(Function &F) {
  // Classic SCoPs.
  for_each(SD->begin(), SD->end(), [&, this] (const Region *R)
  {
    ValidRegions.insert(R);
  });

  if (EnableJitable || EnableExtended) {
    findJitableRegions(F);

    // JITable SCoPs.
    for_each(Regions.begin(), Regions.end(), [&, this] (const Region *R)
    {
      ++(Jitables.TotalJitableRegions);
      ValidRegions.insert(R);
      ++JITScops;
    });
  }
}

/// Iterate over all rejected regions and check the rejected status.
/// If a rejected region is jitable, check the parent region if it
/// would be jitable too. To do this, invoke the SCoPDetection on the
/// parent region and continue...
//
void JITScopDetection::findJitableRegions(Function &F) {
  DepthList depth; // Contains the jitable regions sorted by the tree level.
  DepthElem elem;

  // Stage 1: Create a list of rejected regions grouped by level in the tree.
  depth = initJitableRegionQueue();
  // Stage 2: Check if all siblings are jitable.

  // Level 0 is not the entry region! This way every Region has a Parent.
  for (int i=depth.size()-1; i>=0; i--) {
    elem = depth[i];

    /* Scan the current elment of jitable-regions for merging potential.
     * If all children of one parent are jitable, check the parent and
     * remove children from the current elment if the parent is jitable.
     */
    depth[i] = getNewJitableElem(elem, depth, i);
  }

  // Stage 3: Copy results
  for (DepthList::iterator it = depth.begin(), end = depth.end(); it != end;
       ++it)
  {
    elem = (*it);
    for (DepthElem::iterator eit = elem.begin(), end = elem.end(); eit != end;
         ++eit) {
      Regions.insert((*eit));
    }
  }
}

bool JITScopDetection::runOnFunction(Function &F) {
  RI = &getAnalysis<RegionInfo>();
  SE = &getAnalysis<ScalarEvolution>();
  AA = &getAnalysis<AliasAnalysis>();
  SD = &getAnalysis<ScopDetection>();

  ValidRegions.clear();
  Regions.clear();

  collectCoverage(F);
  collectRejectedInfo(F);

  return false;
}

char JITScopDetection::ID = 0;

INITIALIZE_PASS_BEGIN(JITScopDetection, "polly-jit-scops",
                      "Find JITable SCoPs", false,
                      false)
INITIALIZE_AG_DEPENDENCY(AliasAnalysis)
INITIALIZE_PASS_DEPENDENCY(ScopDetection)
INITIALIZE_PASS_DEPENDENCY(ScalarEvolution)
INITIALIZE_PASS_DEPENDENCY(RegionInfo)
INITIALIZE_PASS_END(JITScopDetection, "polly-jit-scops",
                    "Find JITable SCoPs", false,
                    false)

Pass *polly::createJITScopDetectionPass() {
  return new JITScopDetection();
}
