//===- ScopProfileInfo.cpp - Process ProfileInfo-----------------*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// Use profiling information to calculate interesting numbers for SCoPs.
// Possible numbers are for example:
//   SCoP iterations
//   SCoP coverage (BasicBlocks)
//   SCoP inner loop iterations
// This values become more useful when used in a JIT context, applied on
// rejected SCoP candidates.
//
//===----------------------------------------------------------------------===//
#include "polly/ScopProfileInfo.h"

#include "llvm/ADT/Statistic.h"
#include "llvm/Analysis/AliasAnalysis.h"
#include "llvm/Analysis/RegionPass.h"
#include "llvm/Analysis/RegionInfo.h"
#include "llvm/Analysis/RegionIterator.h"

#include "llvm/Support/CommandLine.h"
#include "llvm/Support/raw_ostream.h"

#include "llvm/Support/Debug.h"
#include "llvm/Module.h"

#include <numeric>
#include <string>

using namespace llvm;
using namespace polly;

static cl::opt<bool>
PrintJitable("print-jitables", cl::desc("Print jitable regions."),
           cl::value_desc("Print jitable regions."),
           cl::init(false));

static cl::opt<bool>
PrintCov("print-coverage", cl::desc("Print jitable regions."),
           cl::value_desc("Print jitable regions."),
           cl::init(false));

static cl::opt<bool>
PrintRejected("print-rejected", cl::desc("Print rejected regions."),
           cl::value_desc("Print jitable regions."),
           cl::init(false));

void ScopProfileInfo::accumulateDynamicRegionInfo(const Region *R) {
  //if (PI->getExecutionCount(R->getEnteringBlock()) <= 0)
  //  SkippedScops.push_back(R);

  double ScopExecutions = accumulate(R->block_begin(), R->block_end(), (double)0,
    [&, this] (double insts, Region::const_block_iterator::pointer node )
     -> double 
    {
      double cnt = PI->getExecutionCount(node);

      return insts + (cnt > 0) ? cnt * node->getInstList().size() : 0;
    });

  DynExecCounts[R->getNameStr()] = ScopExecutions;
}

void ScopProfileInfo::accumulateDynamicFunctionInfo(Function &F) {
  double TotalExecutions = accumulate(F.begin(), F.end(), (double)0,
      [&, this] (double insts, BasicBlock &bb) -> double
      {
         double cnt = PI->getExecutionCount(&bb);
         return (cnt > 0) ? insts + cnt * bb.getInstList().size() : insts;
      });

  double execInstr = accumulate(F.begin(), F.end(),(double)0,
      [&, this] (double insts, BasicBlock &bb) -> double
      {
         double cnt = PI->getExecutionCount(&bb);
         return (cnt > 0) ? insts + bb.getInstList().size() : insts;
      });

  double execBBs = accumulate(F.begin(), F.end(),(double)0,
      [&, this] (double insts, BasicBlock &bb) -> double
      {
        double cnt = PI->getExecutionCount(&bb);
        return (cnt > 0) ? ++insts : insts;
      });

  ExecBasicBlocks = execBBs;
  ExecInstructions = execInstr;
  TotalDynExecutions = TotalExecutions;
}

void ScopProfileInfo::printJitable_Regions(ScopProfileInfo::DepthElem &elem,
                                           int level) {
  // Debug.
  if (PrintJitable && (elem.size() > 0)) {
    outs().changeColor(raw_ostream::MAGENTA);
    outs() << "\nJitable Region subtree:\n";
    outs().indent(4) << "Level [" << level << "]\n";

    std::for_each(elem.begin(), elem.end(), [&, this] (const Region *R)
    {
      bool skipped = PI->getExecutionCount(R->getEntry()) <= 0;
      outs().indent(8) << " ";
      if (skipped)
        outs() << "[S] ";
      outs() << R->getNameStr() << "\n";
    });
    outs().resetColor();
  }
}

void ScopProfileInfo::printRejectedInfo(llvm::raw_ostream &OS) const {
  OS.changeColor(raw_ostream::CYAN);
  for (ScopDetection::rejected_iterator i = SD->rej_begin(),
       n = SD->rej_end(); i != n; ++i) {
    RejectInfo *RI = &((*i).second);

    switch(RI->Reason) {
    case NonAffineAccess:
        OS << "non-aff array: ";
        RI->Failed_LHS->print(OS);
        OS << "\n";
        break;
    case NonAffineCondition:
        OS << "non-aff cond: ";
        RI->Failed_LHS->print(OS);
        OS << " <op> ";
        RI->Failed_RHS->print(OS);
        outs() << "\n";
        break;
    case NonAffineLoopBound:
        OS << "non-aff bound: ";
        RI->Failed_LHS->print(OS);
        break;
    case Alias:
        OS << "Base pointer aliasing found\n";
    case Scalar:
        OS << "Scalar dependency\n";
        break;
    default:
        break;
    }
  }
  OS.resetColor();
}

void ScopProfileInfo::collectRejectedInfo(Function &F) {
  // Accumulate data about rejected regions during SCoP detection.
  RejectInfo *RI;
  for (ScopDetection::rejected_iterator SI = SD->rej_begin(),
       SE = SD->rej_end(); SI != SE; ++SI) {
    RI = &((*SI).second);

    if (RI->Reason == NonAffineLoopBound)
      ++(Rejects.NonAffineLoopBound);

    if (RI->Reason == NonAffineAccess)
      ++(Rejects.NonAffineAccess);

    if (RI->Reason == NonAffineCondition)
      ++(Rejects.NonAffineCondition);

    if (RI->Reason == Unknown)
      ++(Rejects.Unknown);

    ++(Rejects.TotalRejects);
  }
}

void ScopProfileInfo::collectCoverage(Function &F) {
  // Accumulate function-wide data (e.g. InstructionCounts)
  accumulateDynamicFunctionInfo(F);

  for_each(SD->begin(), SD->end(),
           [&, this] (const Region *R)
  {
    BBCounts[R->getNameStr()] = std::accumulate(R->block_begin(), R->block_end(), 0,
                ([] (uint64_t x, Region::const_block_iterator::pointer i)
                 { return ++x; }));

    accumulateDynamicRegionInfo(R);
  });
}

long long ScopProfileInfo::getRegionRuntime(const Function *F,
                                           const Region *R) const
{
  ProfileInfo::TimeKey k = std::make_pair(F, R->getNameStr());
  long long time = 0;
  bool found;

  time = PI->getRegionRuntime(k, &found);
  if (found)
    return time;

  outs() << F->getName() << "," << R->getNameStr()<<"\n";
  PI->dumpRegionRuntime();
  assert(found && "JITScopDetection failed!");
  // If we can't find the current region in the counter table, just sum up
  // all of it's children.
  for (Region::const_iterator i = R->begin(), e = R->end(); i != e; ++i)
    time += getRegionRuntime(F, (*i));

  return time;
}

bool ScopProfileInfo::runOnFunction(Function &F) {
  PI = &getAnalysis<ProfileInfo>();
  RI = &getAnalysis<RegionInfo>();
  SE = &getAnalysis<ScalarEvolution>();
  AA = &getAnalysis<AliasAnalysis>();
  SD = &getAnalysis<JITScopDetection>();

  ExecBasicBlocks = 0;
  ExecInstructions = 0;
  TotalDynExecutions = 0;
  Regions.clear();
  DynExecCounts.clear();

  TotalBasicBlocks += F.getBasicBlockList().size();

  collectCoverage(F);
  collectRejectedInfo(F);

  if (PrintCov)
    print(outs(), F.getParent());
  return false;
}

void ScopProfileInfo::print(raw_ostream &OS, const llvm::Module *M) const {
  // Print info for all valid regions
  for_each(SD->begin(), SD->end(), [&, this] (const Region *R)
  {
    printRegionInfo(OS, R, 0);
  });

  // Print info for rejected regions
  if (PrintRejected) {
    for (ScopDetection::rejected_iterator i = SD->rej_begin(),
         n = SD->rej_end(); i != n; ++i) {
      const Region *R = (*i).first;
      if (SD->find(R) == SD->end()) {
        RejectInfo *RI = &((*i).second);
        printRegionInfo(OS, R, RI);
      }
    }
  }
}

void ScopProfileInfo::printRegionInfo(raw_ostream &OS, const Region *R,
                                      const RejectInfo *RI) const {
  Function *F = R->getEntry()->getParent();
  StringRef FnName = F->getName(),
            RnName = R->getNameStr();

  uint64_t regionTime = getRegionRuntime(F, R);
  uint64_t totalTime = PI->getTotalRuntime();
  double time = regionTime / (double)totalTime;
  OS << FnName;
  OS << format(";%lu",   regionTime);
  OS << format(";%f%%", time * 100);
  if (RI) { OS << ";REJECTED "; }
  else if (Regions.count(R)) { OS << ";VALID (JIT) "; }
  else if (SD->find(R) != SD->end()) { OS << ";VALID "; }
  else { OS << "; ERROR"; }

  if (RI) {
    switch(RI->Reason) {
    case NonAffineAccess:    OS << ";non-aff array: "; break;
    case NonAffineCondition: OS << ";non-aff  cond: "; break;
    case NonAffineLoopBound: OS << ";non-aff bound: "; break;
    case Alias:              OS << ";addr aliasing: "; break;
    case Scalar:             OS << ";scalar depend: "; break;
    default:                 OS << ";unknown: "; break;
    }

    if (RI->Failed_LHS) RI->Failed_LHS->print(OS);
    if (RI->Failed_RHS) {
      OS << " <op> ";
      RI->Failed_RHS->print(OS);
    }
  }
  OS << " " << RnName << " ";
  OS << "\n";
}

// Returns the total number of instructions in the region.
//
double ScopProfileInfo::getInstsInSCoP(const Region *R) const {
  double InstsInSCoP = accumulate(R->block_begin(), R->block_end(), 0,
    [] (double insts, Region::const_block_iterator::pointer node )
     -> double
    {
      return insts + node->getInstList().size();
    });

  return InstsInSCoP;
}

char ScopProfileInfo::ID = 0;

INITIALIZE_PASS_BEGIN(ScopProfileInfo, "polly-pprof-info",
                      "Polly - Merge profiling info with SCoPs", false,
                      false)
INITIALIZE_AG_DEPENDENCY(AliasAnalysis)
INITIALIZE_AG_DEPENDENCY(ProfileInfo)
INITIALIZE_PASS_DEPENDENCY(ScopDetection)
INITIALIZE_PASS_DEPENDENCY(ScalarEvolution)
INITIALIZE_PASS_DEPENDENCY(RegionInfo)
INITIALIZE_PASS_END(ScopProfileInfo, "polly-pprof-info",
                    "Polly - Merge profiling info with SCoPs", false,
                    false)

Pass *polly::createScopProfileInfoPass() {
  return new ScopProfileInfo();
}
