//===- ProfileInfoLoaderPass.cpp - LLVM Pass to load profile info ---------===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// This file implements a concrete implementation of profiling information that
// loads the information from a profile dump file.
//
//===----------------------------------------------------------------------===//
#define DEBUG_TYPE "scop-profile-loader"
#include "polly/LinkAllPasses.h"
#include "polly/ScopProfileInfoLoader.h"
#include "polly/JITScopDetection.h"


#include "llvm/BasicBlock.h"
#include "llvm/InstrTypes.h"
#include "llvm/Module.h"
#include "llvm/Pass.h"
#include "llvm/Analysis/Passes.h"
#include "llvm/Analysis/ProfileInfo.h"
#include "llvm/Analysis/ProfileInfoLoader.h"
#include "llvm/Support/CommandLine.h"
#include "llvm/Support/CFG.h"
#include "llvm/Support/Debug.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Support/Format.h"
#include "llvm/ADT/Statistic.h"
#include "llvm/ADT/SmallSet.h"

#include "llvm/Analysis/RegionPass.h"
#include "llvm/Analysis/RegionInfo.h"
#include "llvm/Analysis/RegionIterator.h"

#include <set>
using namespace llvm;

STATISTIC(NumScopRegionsRead, "The # of region timers read.");

static cl::opt<bool>
PrintSorted("print-sorted", cl::init(false),
                    cl::value_desc("Print sorted counters"),
                    cl::desc("Print sorted counters"));
static cl::opt<std::string>
ScopProfileInfoFilename("scop-profile-info-file", cl::init("llvmprof.out"),
                    cl::value_desc("filename"),
                    cl::desc("Profile file loaded by -profile-loader"));

bool ScopLoader::doInitialization(Module &M) {
  if (Filename.empty())
    Filename = ScopProfileInfoFilename;
  PIL = new ProfileInfoLoader("profile-loader", Filename);

  return false;
}

unsigned ScopLoader::offset = 0;
bool ScopLoader::runOnFunction(Function &F) {
  assert(PIL && "ProfileInformation not available!");
  JSD = &getAnalysis<JITScopDetection>();
  RI = &getAnalysis<RegionInfo>();
  unsigned ReadCount = offset + 0;

  RegionInformation.clear();
  std::vector<long long> RegInfo = PIL->getRawRegionTimings();

  if (RegInfo.size() > 0) {
    std::deque<Region *> ToVisit;
    bool isScop;

    // Just assign the correct number of counters to the function. We will
    // assign the proper region later.
    Region *TopLevel = RI->getTopLevelRegion();

    if (F.getName().str() == "main")
      TotalRuntime = RegInfo[ReadCount];

    ToVisit.push_back(TopLevel);
    while (!ToVisit.empty()) {
      Region *Next = ToVisit.front(); ToVisit.pop_front();

      /* We're only tracking TLR-main and SCoPs */
      isScop = JSD->find(Next) != JSD->end();
      if (isScop || ((F.getName() == "main") && Next->isTopLevelRegion())) {
        TimeKey k = std::make_pair(&F, Next->getNameStr());
        RegionInformation[k] = RegInfo[ReadCount++];
      }

      // enqueue children.
      for (Region::iterator i = Next->begin(), e = Next->end(); i != e; ++i)
        ToVisit.push_back((*i));
    }

    NumScopRegionsRead = ReadCount;
  }

  offset = ReadCount;
  return false;
}

bool ScopLoader::doFinalization(Module &M) {
  std::vector<long long> RegInfo = PIL->getRawRegionTimings();

  if (offset != RegInfo.size()) {
    errs() << "WARNING: PAPI-Region timing profile information is "
           << "inconsistent with the current program!\n"
           << offset << " != " << RegInfo.size();
  }

  return false;
}

char ScopLoader::ID = 0;
INITIALIZE_AG_PASS_BEGIN(ScopLoader, ProfileInfo, "scop-profile-loader",
              "Load profile information from llvmprof.out", false, true, false)
INITIALIZE_AG_DEPENDENCY(AliasAnalysis)
INITIALIZE_PASS_DEPENDENCY(RegionInfo)
INITIALIZE_PASS_DEPENDENCY(JITScopDetection)
INITIALIZE_AG_PASS_END(ScopLoader, ProfileInfo, "scop-profile-loader",
              "Load profile information from llvmprof.out", false, true, false)

Pass *polly::createScopProfileInfoLoaderPass(const std::string &Filename) {
  return new ScopLoader(Filename);
}
Pass *polly::createScopProfileInfoLoaderPass() {
  return new ScopLoader();
}
