//===- polly-prof.cpp - Read in and process llvmprof.out data files --------===//
//
//                      The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// This tools is meant for use with the various LLVM profiling instrumentation
// passes.  It reads in the data file produced by executing an instrumented
// program, and outputs a nice report.
//
//===---------------------------------------------------------------------===//
#define DEBUG_TYPE "scop-profile"
#include "polly/LinkAllPasses.h"

#include "llvm/Support/Debug.h"
#include "polly/ScopProfileInfo.h"

#include "llvm/LLVMContext.h"
#include "llvm/Module.h"
#include "llvm/PassManager.h"
#include "llvm/Bitcode/ReaderWriter.h"
#include "llvm/InstrTypes.h"
#include "llvm/ADT/Statistic.h"
#include "llvm/ADT/Triple.h"
#include "llvm/Analysis/ProfileInfo.h"
#include "llvm/Analysis/ProfileInfoLoader.h"
#include "llvm/Analysis/Passes.h"
#include "llvm/Support/CommandLine.h"
#include "llvm/Support/FormattedStream.h"
#include "llvm/Support/IRReader.h"
#include "llvm/Support/ManagedStatic.h"
#include "llvm/Support/MathExtras.h"
#include "llvm/Support/MemoryBuffer.h"
#include "llvm/Support/PassNameParser.h"
#include "llvm/Support/PrettyStackTrace.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Support/Format.h"
#include "llvm/Support/Signals.h"
#include "llvm/Support/system_error.h"
#include "llvm/Target/TargetData.h"
#include "llvm/Target/TargetLibraryInfo.h"
#include "llvm/Target/TargetMachine.h"
#include "llvm/LinkAllPasses.h"
#include "llvm/LinkAllVMCore.h"

#include <algorithm>
#include <iomanip>
#include <map>
#include <set>

using namespace llvm;
using namespace polly;

#define PRINT_STAT(DESC, VAR) \
  outs() << format(" %-45s - ", (const char *)(DESC)) << VAR << "\n"
#define PRINT_GRP_HEADER(DESC) \
  outs() << ">> " << (const char *)(DESC) << " << \n"
#define PRINT_GRP_ELEM(DESC, VAR) \
  outs().indent(4) << format(" %-41s - ", (const char *)(DESC)) << VAR << "\n"
#define PRINT_STAT_PERCENT(DESC, VAR) \
  outs().indent(4) << format(" %-41s - %.5f %\n", (const char *)(DESC), VAR)
#define MAX_WIDTH 128

namespace {
  static cl::list<const PassInfo*, bool, PassNameParser>
  PassList(cl::desc("Optimizations available:"));

  static cl::opt<std::string>
  BitcodeFile(cl::Positional, cl::desc("<program bitcode file>"),
              cl::Required);

  static cl::opt<std::string>
  ProfileDataFile(cl::Positional, cl::desc("<llvmprof.out file>"),
                  cl::Optional, cl::init("llvmprof.out"));

  static cl::opt<std::string>
  DefaultDataLayout("default-data-layout",
            cl::desc("data layout string to use if not specified by module"),
            cl::value_desc("layout-string"), cl::init(""));

  static cl::opt<bool>
  PrintTimings("print-timings", cl::desc("Print function timings"),
               cl::value_desc("enable timings"), cl::init(false));

  static cl::opt<bool>
  PrintPerFunc("print-per-function",
               cl::desc("Print dynamic info per function"),
               cl::value_desc("print dynamic info per function"),
               cl::init(false));

  static cl::opt<bool>
  PrintRaw("print-raw", cl::desc("Print raw counter information"),
               cl::value_desc("enable timings"), cl::init(true));
  // PairSecondSort - A sorting predicate to sort by the second element of a pair.
  template<class T>
  struct PairSecondSortReverse
    : public std::binary_function<std::pair<T, double>,
                                  std::pair<T, double>, bool> {
    bool operator()(const std::pair<T, double> &LHS,
                    const std::pair<T, double> &RHS) const {
      return LHS.second > RHS.second;
    }
  };

  class ScopProfileInfoPrinter : public FunctionPass {
  private:
    std::vector<std::pair<Function *, uint64_t> > DynInstsTable;
    std::vector<std::pair<Function *, int64_t> > TimeTable;
    std::map<Function *, uint64_t> TotalInstsTable;

    ScopProfileInfo *SP;
    ProfileInfo *PI;
    ProfileInfoLoader &PIL;

    uint64_t TotalInstructions;
    uint64_t TotalInstsExec;
    uint64_t TotalInstsInExecSCoP;
    uint64_t TotalInstsInSCoP;
    uint64_t TotalInstsExecInSCoP;
    uint64_t TotalBBsInSCoP;
    uint64_t TotalBBs;
    uint64_t TotalBBsExec;
    uint64_t TotalBBsInSCoPExec;
    uint64_t ExecInstructions;

    long long TotalRuntime;
    long long TimeSpentInSCoPs;

    void printExecutions();

    void printTimings(Module &M);

    void printRaw();
    void printCodeCoverage();
    void printSkippedSCoPs();
    void printStaticCoverage();
    void printExecutedStaticCoverage();
    void printDynamicCoverage();

    void printDynamicCoveragePerFunc();
  public:
    static char ID;
    explicit ScopProfileInfoPrinter(ProfileInfoLoader &_PIL)
      : FunctionPass(ID), PIL(_PIL) {
      TotalInstsInExecSCoP = 0;
      TotalBBs = 0;
      TotalBBsInSCoPExec = 0;
      TotalBBsInSCoP = 0;
      TotalInstsExec = 0;
      TotalInstsInSCoP = 0;
      TotalInstsExecInSCoP = 0;
      TotalBBsExec = 0;
      TotalInstructions = 0;
      ExecInstructions = 0;
      TotalRuntime = 0;

      TimeSpentInSCoPs = 0;
    }

    virtual void getAnalysisUsage(AnalysisUsage &AU) const {
      AU.setPreservesAll();
      AU.addRequiredTransitive<ProfileInfo>();
      AU.addRequiredTransitive<ScopProfileInfo>();
    }

    virtual bool runOnFunction(Function &F);
    virtual bool doFinalization(Module &M);
  };
}

char ScopProfileInfoPrinter::ID = 0;
void ScopProfileInfoPrinter::printRaw() {
  PRINT_GRP_HEADER("Instruction counts");
  PRINT_GRP_ELEM("Instructions in SCoP", TotalInstsInSCoP);
  PRINT_GRP_ELEM("Instructions in executed SCoP", TotalInstsInExecSCoP);
  PRINT_GRP_ELEM("Instructions executed", ExecInstructions);
  PRINT_GRP_ELEM("Instructions exec'd in SCoP", TotalInstsExecInSCoP);
  PRINT_GRP_ELEM("Total instructions executed", TotalInstsExec);
  PRINT_GRP_ELEM("Total instruction count", TotalInstructions);

  PRINT_GRP_HEADER("BasicBlock counts");
  PRINT_GRP_ELEM("BasicBlocks in SCoP", TotalBBsInSCoP);
  PRINT_GRP_ELEM("BasicBlocks (total)", TotalBBs);
  PRINT_GRP_ELEM("BasicBlocks exec'd (total)", TotalBBsExec);
  PRINT_GRP_ELEM("BasicBlocks exec'd in SCoP", TotalBBsInSCoPExec);

  PRINT_GRP_HEADER("Rejects");
  PRINT_GRP_ELEM("Rejected regions", SP->getRejects().TotalRejects);
  PRINT_GRP_ELEM("Non affine loop bounds", SP->getRejects().NonAffineLoopBound);
  PRINT_GRP_ELEM("Non affine access", SP->getRejects().NonAffineAccess);
  PRINT_GRP_ELEM("Non affine condition", SP->getRejects().NonAffineCondition);
  PRINT_GRP_ELEM("Unknown", SP->getRejects().Unknown);

  PRINT_GRP_HEADER("Timing");
  PRINT_GRP_ELEM("Time Spent in SCoPs [us]", TimeSpentInSCoPs);
  PRINT_GRP_ELEM("Total run-time [us]", TotalRuntime);

  PRINT_GRP_HEADER("JIT");
  PRINT_GRP_ELEM("Jitable regions", SP->getJitables().TotalJitableRegions);
}

void ScopProfileInfoPrinter::printCodeCoverage() {
  double coverage = 0;

  coverage = ExecInstructions / (double)TotalInstructions * 100;
  PRINT_STAT_PERCENT("Code coverage", coverage);
}

void ScopProfileInfoPrinter::printSkippedSCoPs() {
  uint64_t skippedScops = SP->getSkippedSCoPs();
  PRINT_GRP_ELEM("Skipped SCoPs", skippedScops);
}

void ScopProfileInfoPrinter::printExecutedStaticCoverage() {
  double coverage = 0;

  coverage = TotalInstsInExecSCoP / (double)TotalInstructions * 100;
  PRINT_STAT_PERCENT("Static SCoP coverage (exec'd)", coverage);
}

void ScopProfileInfoPrinter::printStaticCoverage() {
  double coverage = 0;

  coverage = TotalInstsInSCoP / (double)TotalInstructions * 100;
  PRINT_STAT_PERCENT("Static SCoP coverage", coverage);
}

void ScopProfileInfoPrinter::printDynamicCoverage() {
  double coverage = 0;

  // Instruction based (deprecated)
  //coverage = TotalInstsExecInSCoP / (double)TotalInstsExec * 100;
  //PRINT_STAT_PERCENT("Dynamic SCoP coverage", coverage);
  coverage = (double)TimeSpentInSCoPs / (double)TotalRuntime * 100;
  PRINT_STAT_PERCENT("Dynamic SCoP coverage", coverage);
}

void ScopProfileInfoPrinter::printDynamicCoveragePerFunc() {
  sort(DynInstsTable.begin(), DynInstsTable.end(),
       PairSecondSortReverse<Function *>());

  outs() << "\n===" << std::string(73, '-') << "===\n";
  outs() << "Dynamic SCoP coverage (per function):\n\n";

  outs() << " #   DynCov    Runtime  Name\n";
  int sz = DynInstsTable.size();
  for (int64_t i = 0, e = ((sz < 10) ? sz : 10); i != e; ++i) {
    Function *F = DynInstsTable[i].first;
    uint64_t dynInsts = DynInstsTable[i].second;

    if (TotalInstsTable[F] == 0) {
      outs() << "\n  NOTE: " << e-i << " function"
        << (e-i-1 ? "s were" : " was") << " never executed!\n";
      break;
    }
    // No. DynInsts  DynCov [%] Runtime [%] Name
    outs() << format("%2ld. %8.3f%% %8.3f%% %s\n",
                     i+1,
                     dynInsts / (double) TotalInstsTable[F] * 100,
                     PI->getRuntime(F) / (double) TotalRuntime * 100,
                     DynInstsTable[i].first->getName());
  }
  outs() << "\n";
}

bool ScopProfileInfoPrinter::doFinalization(Module &M) {
  TotalRuntime = (double)PI->getTotalRuntime();
  printExecutions();

  if (PrintRaw) {
    printRaw();
    PRINT_GRP_HEADER("Coverage");
    printStaticCoverage();
    printSkippedSCoPs();
    printExecutedStaticCoverage();
    printCodeCoverage();
  }

  printDynamicCoverage();

  if (PrintPerFunc)
    printDynamicCoveragePerFunc();

  if (PrintTimings)
    printTimings(M);

  return false;
}

void ScopProfileInfoPrinter::printExecutions() {
  outs() << "LLVM profiling output for execution";
  if (PIL.getNumExecutions() != 1) outs() << "s";
  outs() << ":\n";

  for (int64_t i = 0, e = PIL.getNumExecutions(); i != e; ++i) {
    outs() << "  ";
    if (e != 1) outs() << i+1 << ". ";
    outs() << PIL.getExecution(i) << "\n";
  }

  outs() << "===" << std::string(73, '-') << "===\n\n";
}

void ScopProfileInfoPrinter::printTimings(Module &M) {
  outs() << "\n";
  sort(TimeTable.begin(), TimeTable.end(), PairSecondSortReverse<Function *>());

  outs() << "\n===" << std::string(73, '-') << "===\n";
  outs() << "Function execution time:\n\n";

  outs() << " #   Runtime    Runtime[%]  Calls[#]  Name\n";
  int sz = TimeTable.size();
  for (int64_t i = 0, e = ((sz < 10) ? sz : 10); i != e; ++i) {
    if (TimeTable[i].second == 0) {
      outs() << "\n  NOTE: " << e-i << " function"
        << (e-i-1 ? "s were" : " was") << " never executed!\n";
      break;
    }

    outs() << format("%2ld. %8.5fs %8.5f%% %d %s\n",
                     i+1, TimeTable[i].second / (double)1000000,
                     100 * (TimeTable[i].second / (double)TotalRuntime),
                     (unsigned)PI->getExecutionCount(TimeTable[i].first),
                     TimeTable[i].first->getName());
  }
  outs() << "\n";
}

bool ScopProfileInfoPrinter::runOnFunction(Function &F) {
  SP = &getAnalysis<ScopProfileInfo>();
  PI = &getAnalysis<ProfileInfo>();

  uint64_t dynInstsPerFunc = 0;
  uint64_t totalInstsPerFunc = 0;
  long long timeSpentInSCoP = 0;

  TotalBBs         += F.getBasicBlockList().size();
  TotalBBsExec     += SP->getExecutedBasicBlockCount();
  ExecInstructions += SP->getExecInstructions();
  totalInstsPerFunc = SP->getTotalExecutions();

  for (Function::iterator FI = F.begin(), FE = F.end(); FI != FE; ++FI)
    TotalInstructions += FI->getInstList().size();

  for (ScopProfileInfo::iterator SPI = SP->begin(), SPE = SP->end();
       SPI != SPE; ++SPI) {
    TotalBBsInSCoP        += SP->getBasicBlockCount(*SPI);
    TotalBBsInSCoPExec    += SP->getBasicBlockCountIfExecuted(*SPI);
    TotalInstsInSCoP      += SP->getInstsInSCoP(*SPI);
    TotalInstsInExecSCoP  += SP->getInstsInExecSCoP(*SPI);

    dynInstsPerFunc       += SP->getDynExecutions(*SPI);

    timeSpentInSCoP       += SP->getRegionRuntime(&F, (*SPI));
  }

  TimeTable.push_back(std::make_pair(&F, PI->getRuntime(&F)));
  DynInstsTable.push_back(std::make_pair(&F, dynInstsPerFunc));

  TotalInstsTable[&F] = totalInstsPerFunc;
  TotalInstsExecInSCoP  += dynInstsPerFunc;
  TotalInstsExec        += totalInstsPerFunc;
  TimeSpentInSCoPs      += timeSpentInSCoP;

  return false;
}

int main(int argc, char **argv) {
  // Print a stack trace if we signal out.
  sys::PrintStackTraceOnErrorSignal();
  PrettyStackTraceProgram X(argc, argv);

  llvm_shutdown_obj Y;  // Call llvm_shutdown() on exit.
  LLVMContext &Context = getGlobalContext();

  PassRegistry &Registry = *PassRegistry::getPassRegistry();
  initializeCore(Registry);
  initializeScalarOpts(Registry);
  initializeIPO(Registry);
  initializeAnalysis(Registry);
  initializeIPA(Registry);
  initializeTransformUtils(Registry);
  initializeInstCombine(Registry);
  initializeInstrumentation(Registry);
  initializeTarget(Registry);
  initializeCodePreparationPass(Registry);
  initializeScopDetectionPass(Registry);
  initializeScopProfileInfoPass(Registry);

  cl::ParseCommandLineOptions(argc, argv,
                              "polly - SCoP profiler\n");

  // Allocate a full target machine description only if necessary.
  // FIXME: The choice of target should be controllable on the command line.
  std::auto_ptr<TargetMachine> target;

  SMDiagnostic Err;

  // Load the input module...
  std::auto_ptr<Module> M;
  M.reset(ParseIRFile(BitcodeFile, Err, Context));

  if (M.get() == 0) {
    Err.print(argv[0], errs());
    return 1;
  }

  PassManager Passes;

  // Add an appropriate TargetData instance for this module.
  TargetData *TD = 0;
  const std::string &ModuleDataLayout = M.get()->getDataLayout();
  if (!ModuleDataLayout.empty())
    TD = new TargetData(ModuleDataLayout);
  else if (!DefaultDataLayout.empty())
    TD = new TargetData(DefaultDataLayout);

  if (TD)
    Passes.add(TD);

  for (unsigned i=0; i < PassList.size(); ++i) {
    const PassInfo *PassInf = PassList[i];
    Pass *P = 0;
    if (PassInf->getNormalCtor())
      P = PassInf->getNormalCtor()();
    else
      errs() << argv[0] << ": cannot create pass: "
             << PassInf->getPassName() << "\n";
    if (P)
      Passes.add(P);
  }

  std::string Name = M->getModuleIdentifier();
  size_t pos = Name.find(".linked.ll");

  Name = Name.substr(0, pos);

  outs() << "---------------------------------------------------------------\n";
  outs() << ">>> ========= '" << Name << "' Program\n";
  outs() << "---------------------------------------------------------------\n";

  // Read the profiling information. This is redundant since we load it again
  // using the standard profile info provider pass, but for now this gives us
  // access to additional information not exposed via the ProfileInfo
  // interface.
  ProfileInfoLoader PIL(argv[0], ProfileDataFile);

  Passes.add(createScopProfileInfoLoaderPass(ProfileDataFile));
  Passes.add(createScopProfileInfoPass());
  Passes.add(new ScopProfileInfoPrinter(PIL));
  Passes.run(*M.get());

  return 0;
}
